﻿/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2022, Kassow Robots
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Kassow Robots nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

#include "kr2/kord/kord_server.h"
#include "kr2/kord/robot_model.h"
#include "kr2/kord/robot_controller.h"
#include "kr2/kord/robot_control_action.h"

#include "kr2/kord/protocol/KORDFrames.h"
#include "kr2/kord/protocol/StatusFrameBuilder.h"
#include "kr2/kord/protocol/ContentFrameBuilder.h"
#include "kr2/kord/protocol/ContentFrameParser.h"
#include "kr2/kord/protocol/DataFormatDescription.h"

#include <kr2_rc_api/api_v1/arm/trajectory.h>

#include <kr2/utils/stats.h>

#include <kr2/utils/counter.h>
#include <kr2/utils/socket.h>

#include <sys/types.h>
#include <cstdlib>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <cstdio>
#include <cstring>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

// RT capability
#include <csignal>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <mutex>
#include <thread>

#include <atomic>
#include <memory>
#include <chrono>
#include <kr2/kord/system/RobotControllerFlags.h>

using namespace kr2::kord;
using namespace kr2::kord::protocol;
using namespace std::chrono_literals;

typedef std::chrono::time_point<std::chrono::high_resolution_clock> timepoint;

class KORDServer::KORDServerImpl{
public:
    KORDServerImpl(int a_port, int a_session_id, kr2::kord::EConnType a_conn_type, unsigned int a_tx_freq):
        port_(a_port),
        session_id_(a_session_id),
        type_(a_conn_type),
        freq_(a_tx_freq)
    {
        resetAllFrames();
    }

    bool open(double timeout);
    bool close();

    void capture();

    int send(uint8_t *buffer, size_t dlen);

    unsigned int latestFrameID();

    void updateArmStatusStamps(int64_t& sent_status_tx_stamp, CommandsMonitor& stats_commands);
    void updateKordFrame();

    int transmitFrame();

    bool offsetValid(EKORDDataID) const;
    bool updateArrayData(EKORDDataID, const std::array<uint16_t, 7UL>&);
    bool updateArrayData(EKORDDataID, const std::array<unsigned int, 7UL>&);
    bool updateArrayData(EKORDDataID, const std::array<double, 7UL>&);
    bool updateArrayData(EKORDDataID, const std::array<double, 6UL>&);
    bool updateArrayData(EKORDDataID, const std::array<double, 3UL>&);
    bool updateVectorData(EKORDDataID, const std::vector<double>&);

    void resetAllFrames(){
        capture_frm_.resetCaptureFrames();
        transmit_frm_.resetTransmitFrames();
    }

public:

    CaptureFRM capture_frm_;
    TransmitFRM transmit_frm_;

public:
//    StampsLog ts_record; ts_record.reset();
    unsigned int port_;
    unsigned int session_id_; // KORD Session ID (comes from TP)
    EConnType type_;
    unsigned int freq_;

    // [AK:Future] Have a one instance of socket
    // Socket *masterSocket = nullptr; // will be initialized in *open* method
    
    int sock_, length_;
    socklen_t fromlen_;
    struct sockaddr_in server_, from_;

    Counter thread_counter;

    std::thread capture_;
    std::mutex mux_capture_;
    std::mutex mux_transmit_;
    std::mutex mux_stats_update;
    std::atomic<bool> keep_running_;
    DataFormatDescription status_dfd = DataFormatDescription::makeStatusFrameDescription(0);

    // Iterative update
    int frame_index_ = 0;
    int load_index_ = 0;
    int cpu_state_index = 0;
};

bool KORDServer::KORDServerImpl::open(double a_timeout)
{
    // [AK:Future] How about this change for the socket?
    // int status;
    // std::string ip  = "192.168.39.1";
    // std::string port = "7582";

    // // Socket *masterSocket = new Socket(AF_INET,SOCK_STREAM,0); //AF_INET (Internet mode) SOCK_STREAM (TCP mode) 0 (Protocol any)
    // Socket *masterSocket = new Socket(AF_INET,SOCK_DGRAM,0); //AF_INET (Internet mode) SOCK_STREAM (UDP mode) 0 (Protocol any)

    // // int optVal = 1;
    // timeval optVal; optVal.tv_sec = 1; optVal.tv_usec = 0;
    // status = masterSocket->socket_set_opt_new(SOL_SOCKET, SO_RCVTIMEO, optVal); //You can reuse the address and the port
    // if (status < 0 ) 
    // {
    //     CLOG_INFO("Opening socket failed");
    //     return false;
    // }

    // status = masterSocket->bind(ip, port); //Bind socket on localhost:1234
    // if (status < 0) 
    // {
    //     CLOG_INFO("Failed to bind\n");
    //     return false;
    // }

    //
    sock_=socket(AF_INET, SOCK_DGRAM, 0);
    if (sock_ < 0) {
        CLOG_INFO("Opening socket failed");
        return false;
    }

    // receive timeout of 1 second is set
    timeval read_timeout;
    read_timeout.tv_sec = 1;
    read_timeout.tv_usec = 0;
    setsockopt(sock_, SOL_SOCKET, SO_RCVTIMEO, &read_timeout, sizeof(read_timeout));

    length_ = sizeof(server_);
    memset(&server_, 0x00, length_);
    server_.sin_family=AF_INET;
    server_.sin_addr.s_addr=INADDR_ANY;
    server_.sin_port=htons(port_);
    if (::bind(sock_,(struct sockaddr *)&server_,length_)<0){
        CLOG_INFO("Failed to bind\n");
        return 0;
    }
    fromlen_ = sizeof(struct sockaddr_in);

    // thread_counter.set_sock(sock_);
    // thread_counter.set_from(from_);
    // thread_counter.set_fromlen(fromlen_);
    // thread_counter.set_capture_frm(capture_frm_);
    // thread_counter.start();

    keep_running_.store(true);
    capture_ = std::thread(
        [this](){capture();}
    );

    return true;
}

bool KORDServer::KORDServerImpl::close()
{
    keep_running_.store(false);
    // thread_counter.stop();

    if (capture_.joinable()){
        capture_.join();
    }

    // if (thread_counter.joinable()){
    //     thread_counter.join();
    // }
    
    // masterSocket->close();
    int n = ::close(sock_);
    // CLOG_INFO("Socket close error: " << std::to_string(n));
    // std::cout << std::flush;

    // return (masterSocket->is_active());
    return (n != 0);
}

// Should get the **capture_frm_** for the change
// keep_running_ --> is a flag to terminate
void KORDServer::KORDServerImpl::capture() {
    
    // output of socket 
    int n;

    // temporary variables
    std::chrono::high_resolution_clock::time_point current_time;
    
    while (keep_running_.load()) {
        // This is a draft for communication with *masterSocket*
        // kr2::kord::KORDFrame buf;
        // int len = static_cast<int>(sizeof(kr2::kord::KORDFrame));
        // std::string ip  = "192.168.39.1";
        // std::string port = "7582";
        // n = masterSocket->socket_readFrom(buf, len, ip, port);
        
        // receive and check n_bytes as an input
        n = recvfrom(sock_, &capture_frm_.kord_frame, sizeof(kr2::kord::protocol::KORDFrame), 0, (struct sockaddr *)&from_, &fromlen_);
        if (n < 0)  { boost::interprocess::scoped_lock lock(mux_capture_); capture_frm_.fail_to_read_error++; continue; }
        if (n == 0) { boost::interprocess::scoped_lock lock(mux_capture_); capture_frm_.fail_to_read_empty++; continue; }

        current_time    = std::chrono::high_resolution_clock::now();
        mux_stats_update.lock();
        capture_frm_.recent_captured_timestamp = current_time;
        mux_stats_update.unlock();

        auto frame_id__ = ntohs(capture_frm_.kord_frame.frame_id_);

        // FRAME --> CONTENT
        if (frame_id__ == kr2::kord::protocol::KORD_FRAME_ID_CONTENT) 
        {
            mux_capture_.lock();
            capture_frm_.frame_id = frame_id__;
            capture_frm_.content_parser.setFromPayload(capture_frm_.kord_frame.payload_,  ntohs(capture_frm_.kord_frame.payload_length_));
            auto capture_frm_items_num = capture_frm_.content_parser.getItemsCount();

            // statistics
            //get current timespec
            mux_stats_update.lock();
            capture_frm_.recent_captured_content_timestamp = kr2::utils::chronoToTimespec(current_time);
            mux_stats_update.unlock();

            for (int idx = 0; idx < capture_frm_items_num; idx++)
            {
                // filling capture_frm_
                kr2::kord::protocol::EKORDItemID iid = capture_frm_.content_parser.getItemID(idx);
                capture_frm_.iid = iid;
                ContentItem item = capture_frm_.content_parser.getItemContent(idx);
                capture_frm_.captureIIDInfo(iid, item);
            }
            mux_capture_.unlock();
        }

        // FRAME --> ID_STATUS
        if (frame_id__ == kr2::kord::protocol::KORD_FRAME_ID_STATUS) { CLOG_INFO("[KORD_CAPTURE] Status captured, not processing..."); }
    }
}

int KORDServer::KORDServerImpl::send(uint8_t *buffer, size_t dlen)
{
    // std::string ip  = "192.168.39.1";
    // std::string port = "7582";
    // return masterSocket->socket_writeTo(buffer, dlen, ip, port);
    return sendto(sock_, buffer, dlen,0, (struct sockaddr *)&from_, fromlen_);
}

unsigned int KORDServer::KORDServerImpl::latestFrameID()
{
    mux_capture_.lock();
    unsigned int fid = capture_frm_.frame_id;
    mux_capture_.unlock();
    return fid;
}

void KORDServer::KORDServerImpl::updateArmStatusStamps(int64_t& sent_status_tx_stamp, CommandsMonitor& stats_monitor)
{
    int64_t     ts;
    uint32_t    fail_to_read_error;
    uint32_t    fail_to_read_empty;

    Ana         ltick_ana;
    Ana         stick_ana;
    Ana wake_up_ana;
    int64_t local_packets_lost_seq;
    int64_t local_packets_lost_timestmp;
    int64_t global_ltick_max; // command jitter global MAX
    int64_t global_packets_lost_seq; // commands lost global counter
    int64_t global_packets_lost_timestmp;
    int64_t global_system_jitter_max; // system jitter global MAX
    int64_t global_stick_max = 0; // global round trip time max

    uint16_t    max_frames_in_tick;
    int64_t     faulty_frames_start;

    mux_capture_.lock();
    
    ts                  = capture_frm_.status_req.getData<int64_t>(capture_frm_.dfd_arm_status_req.getOffset(EKORDDataID::eTxStamp));
    fail_to_read_empty  = capture_frm_.fail_to_read_empty;
    fail_to_read_error  = capture_frm_.fail_to_read_error;
    faulty_frames_start = capture_frm_.faulty_frames_start;
    max_frames_in_tick  = capture_frm_.max_frames_in_tick;

    mux_capture_.unlock();

    mux_stats_update.lock();

    ltick_ana.set(stats_monitor.accum_ltick.get_ana());
    stick_ana.set(stats_monitor.accum_stick.get_ana());
    wake_up_ana.set(stats_monitor.accum_wake_up_stats.get_ana());

    local_packets_lost_seq = stats_monitor.local_packets_lost_seq;
    local_packets_lost_timestmp = stats_monitor.local_packets_lost_timestmp;

    global_ltick_max = stats_monitor.global_ltick_max;
    global_packets_lost_seq = stats_monitor.global_packets_lost_seq;
    global_packets_lost_timestmp = stats_monitor.global_packets_lost_timestmp;
    global_system_jitter_max = stats_monitor.global_system_jitter_max;
    global_stick_max = stats_monitor.global_stick_max;
    mux_stats_update.unlock();

    mux_transmit_.lock();
    transmit_frm_.status_builder.addSequenceNumber(transmit_frm_.seq_num_++);
    
    transmit_frm_.status_builder.addData<int64_t>(ts, status_dfd.getOffset(EKORDDataID::eTxStampEcho));
    auto time_stamp_ = std::chrono::high_resolution_clock::now().time_since_epoch().count();
    transmit_frm_.status_builder.addData<int64_t>(
        time_stamp_,
        status_dfd.getOffset(EKORDDataID::eTxStamp)
    );
    sent_status_tx_stamp = time_stamp_;

    transmit_frm_.status_builder.addData(fail_to_read_error, status_dfd.getOffset(EKORDDataID::eFailToReadError));

    transmit_frm_.status_builder.addData(fail_to_read_empty, status_dfd.getOffset(EKORDDataID::eFailToReadEmpty));

    // old variant
    // int64_t min_delay = ltick_ana.min;
    // transmit_frm_.status_builder.addData(min_delay, status_dfd.getOffset(EKORDDataID::eMinDelay));

    // int64_t max_delay = ltick_ana.max;
    // transmit_frm_.status_builder.addData(max_delay, status_dfd.getOffset(EKORDDataID::eMaxDelay));
    
    // int64_t avg_delay = ltick_ana.avg;
    // transmit_frm_.status_builder.addData(avg_delay, status_dfd.getOffset(EKORDDataID::eAverageDelay));

    // int64_t min_tick_delay = stick_ana.min;
    // transmit_frm_.status_builder.addData(min_tick_delay, status_dfd.getOffset(EKORDDataID::eMinTickDelay));
    
    // int64_t max_tick_delay = stick_ana.max;
    // transmit_frm_.status_builder.addData(max_tick_delay, status_dfd.getOffset(EKORDDataID::eMaxTickDelay));

    // int64_t avg_tick_delay = stick_ana.avg;
    // transmit_frm_.status_builder.addData(avg_tick_delay, status_dfd.getOffset(EKORDDataID::eAverageTickDelay));

    // new variant
    int64_t cmd_jitter_window_max = ltick_ana.max;
    transmit_frm_.status_builder.addData(cmd_jitter_window_max, status_dfd.getOffset(EKORDDataID::eMaxCmdJitterWindow));

    int64_t cmd_jitter_window_avg = ltick_ana.avg;
    transmit_frm_.status_builder.addData(cmd_jitter_window_avg, status_dfd.getOffset(EKORDDataID::eAvgCmdJitterWindow));

    int64_t cmd_jitter_global_max = global_ltick_max;
    transmit_frm_.status_builder.addData(cmd_jitter_global_max, status_dfd.getOffset(EKORDDataID::eMaxCmdJitterGlobal));

    int64_t round_trip_window_max = stick_ana.max;
    transmit_frm_.status_builder.addData(round_trip_window_max, status_dfd.getOffset(EKORDDataID::eMaxTickDelayWindow));

    int64_t round_trip_window_avg = stick_ana.avg;
    transmit_frm_.status_builder.addData(round_trip_window_avg, status_dfd.getOffset(EKORDDataID::eAvgTickDelayWindow));

    int64_t round_trip_global_max = global_stick_max;
    transmit_frm_.status_builder.addData(round_trip_global_max, status_dfd.getOffset(EKORDDataID::eMaxTickDelayGlobal));

    int64_t cmd_lost_window_seq = local_packets_lost_seq;
    transmit_frm_.status_builder.addData(cmd_lost_window_seq, status_dfd.getOffset(EKORDDataID::eCmdLostWindowSeq));

    int64_t cmd_lost_window_timestmp = local_packets_lost_timestmp;
    transmit_frm_.status_builder.addData(cmd_lost_window_timestmp, status_dfd.getOffset(EKORDDataID::eCmdLostWindowTimestamp));

    int64_t cmd_lost_global_seq = global_packets_lost_seq;
    transmit_frm_.status_builder.addData(cmd_lost_global_seq, status_dfd.getOffset(EKORDDataID::eCmdLostGlobalSeq));

    int64_t cmd_lost_global_timestmp = global_packets_lost_timestmp;
    transmit_frm_.status_builder.addData(cmd_lost_global_timestmp, status_dfd.getOffset(EKORDDataID::eCmdLostGlobalSeq));

    int64_t system_jitter_window_max = wake_up_ana.max;
    transmit_frm_.status_builder.addData(system_jitter_window_max, status_dfd.getOffset(EKORDDataID::eMaxSysJitterWindow));

    int64_t system_jitter_window_avg = wake_up_ana.avg;
    transmit_frm_.status_builder.addData(system_jitter_window_avg, status_dfd.getOffset(EKORDDataID::eAvgSysJitterWindow));

    int64_t system_jitter_global_max = global_system_jitter_max;
    transmit_frm_.status_builder.addData(system_jitter_global_max, status_dfd.getOffset(EKORDDataID::eMaxSysJitterGlobal));

    mux_transmit_.unlock();
}

void KORDServer::KORDServerImpl::updateKordFrame()
{
    mux_transmit_.lock();
    transmit_frm_.kord_frame.frame_id_ = htons(kr2::kord::protocol::KORD_FRAME_ID_STATUS);
    transmit_frm_.kord_frame.session_id_ = htons(session_id_);
    transmit_frm_.kord_frame.tx_timestamp_ = std::chrono::high_resolution_clock::now().time_since_epoch().count();
    transmit_frm_.kord_frame.payload_length_ = htons(sizeof(kr2::kord::protocol::KORDStatusFrame));
    transmit_frm_.kord_frame.kord_version_ = htons(KORD_PROTOCOL_VERSION);
    transmit_frm_.status_builder.getPayload(transmit_frm_.kord_frame.payload_, sizeof(transmit_frm_.kord_frame.payload_));
    mux_transmit_.unlock();
}

int KORDServer::KORDServerImpl::transmitFrame()
{
    return send(reinterpret_cast<uint8_t*>(&transmit_frm_.kord_frame), sizeof(KORDFrame));
}

bool KORDServer::KORDServerImpl::offsetValid(EKORDDataID a_iid) const
{
    unsigned int offset = status_dfd.getOffset(a_iid);

    // Check if the KORD Item ID exists in the description
    if (offset == UINT32_MAX){
        return false;
    }

    return true;
}

bool KORDServer::KORDServerImpl::updateArrayData(EKORDDataID a_iid, const std::array<uint16_t, 7UL> &a_arr)
{
    unsigned int offset = status_dfd.getOffset(a_iid);

    // Check if the KORD Item ID exists in the description
    if (offset == UINT32_MAX){
        return false;
    }

    transmit_frm_.status_builder.addData<std::array<uint16_t, 7UL>>(a_arr, offset);
    return true;
}


bool KORDServer::KORDServerImpl::updateArrayData(EKORDDataID a_iid, const std::array<unsigned int, 7UL> &a_arr)
{
    unsigned int offset = status_dfd.getOffset(a_iid);

    // Check if the KORD Item ID exists in the description
    if (offset == UINT32_MAX){
        return false;
    }

    transmit_frm_.status_builder.addData<std::array<unsigned int, 7UL>>(a_arr, offset);
    return true;
}

bool KORDServer::KORDServerImpl::updateArrayData(EKORDDataID a_iid, const std::array<double, 6UL> &a_arr)
{
    unsigned int offset = status_dfd.getOffset(a_iid);

    // Check if the KORD Item ID exists in the description
    if (offset == UINT32_MAX){
        return false;
    }

    transmit_frm_.status_builder.addData<std::array<double, 6UL>>(a_arr, offset);
    return true;
}

bool KORDServer::KORDServerImpl::updateArrayData(EKORDDataID a_iid, const std::array<double, 7UL> &a_arr)
{
    unsigned int offset = status_dfd.getOffset(a_iid);

    // Check if the KORD Item ID exists in the description
    if (offset == UINT32_MAX){
        return false;
    }

    transmit_frm_.status_builder.addData<std::array<double, 7UL>>(a_arr, offset);
    return true;
}

bool KORDServer::KORDServerImpl::updateArrayData(EKORDDataID a_iid, const std::array<double, 3UL> &a_arr)
{
    unsigned int offset = status_dfd.getOffset(a_iid);

    // Check if the KORD Item ID exists in the description
    if (offset == UINT32_MAX){
        return false;
    }

    transmit_frm_.status_builder.addData<std::array<double, 3UL>>(a_arr, offset);
    return true;
}

bool KORDServer::KORDServerImpl::updateVectorData(EKORDDataID a_iid, const std::vector<double>& a_data)
{
    if (a_data.size() == 7){
        std::array<double, 7UL> ar;
        std::copy(a_data.begin(), a_data.end(), ar.begin());
        updateArrayData(a_iid, ar);
        return true;
    }
    if (a_data.size() == 6){
        std::array<double, 6UL> ar;
        std::copy(a_data.begin(), a_data.end(), ar.begin());
        updateArrayData(a_iid, ar);
        return true;
    }
    if (a_data.size() == 3){
        std::array<double, 3UL> ar;
        std::copy(a_data.begin(), a_data.end(), ar.begin());
        updateArrayData(a_iid, ar);
        return true;
    }

    return false;
}

kr2::kord::KORDServer::KORDServer(int port, int session_id, kr2::kord::EConnType conn_type, unsigned int tx_freq):
    impl_(new KORDServerImpl(port, session_id, conn_type, tx_freq))
{ }

kr2::kord::KORDServer::~KORDServer()
{
    delete impl_;
}

bool kr2::kord::KORDServer::open(double a_timeout)
{
    return impl_->open(a_timeout);
}

bool kr2::kord::KORDServer::isActive() const
{
    // return (impl_->masterSocket->is_active());
    return (impl_->sock_ > 0);
}

bool kr2::kord::KORDServer::close()
{
//    if (use_rt_){
//        close(latency_target_fd);
//    }

    return impl_->close();
}

std::vector<kr2::kord::KORDItemId> kr2::kord::KORDServer::fetchItemsIDs() const
{
    std::vector<kr2::kord::KORDItemId> items;

    items.push_back(impl_->latestFrameID());

    return items;
}

bool kr2::kord::KORDServer::fetchItem(kr2::kord::RobotControl *a_control, kr2::kord::tick_t a_recent, CommandsMonitor& stats_commands, const RobotController &controller)
{
    using TTP =  kr2rc_api::Trajectory::CmdTrajectoryParams;

    auto ad = std::chrono::seconds(a_recent.tv_sec)
            + std::chrono::nanoseconds(a_recent.tv_nsec);
    timepoint latest_ok = timepoint{std::chrono::duration_cast<std::chrono::high_resolution_clock::duration>(ad)};

    impl_->mux_capture_.lock();
    kr2::kord::protocol::EKORDItemID iid;
    iid = impl_->capture_frm_.iid;
    a_control->item_id_ = iid;

    a_control->remote_host_ = impl_->from_; //  MUST BE PASSED!
    // a_control->remote_host_ = (struct sockaddr *)&impl_->masterSocket->address_info.ai_addr;

    switch(iid){
        case kr2::kord::protocol::EKORDItemID::eCommandMoveJ:{
            if (!stats_commands.getAlarmState() && (controller.motion_flags_ & kr2::kord::protocol::EMotionFlags::MOTION_FLAG_SUSPENDED)){ // suspended 
                stats_commands.setAlarmState(true);
            }
            if (stats_commands.getAlarmState()){ // if alarm - ignore commands = clearing old capture frame
                impl_->capture_frm_.iid = EKORDItemID::eNone;
                break;
            }
            a_control->capture_ts_ = impl_->capture_frm_.rcv_cmd_joints_ts;
            a_control->sequence_num_ = impl_->capture_frm_.move_J.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_move_J.getOffset(EKORDDataID::eSequenceNumber));
            a_control->move_j_.setTargets(impl_->capture_frm_.move_J.getData<std::array<double, 7UL>>           (impl_->capture_frm_.dfd_cmd_move_J.getOffset(EKORDDataID::eJConfigurationArm)));
            a_control->move_j_.setTrackingType((TTP::TrackingType)impl_->capture_frm_.move_J.getData<u_int8_t>  (impl_->capture_frm_.dfd_cmd_move_J.getOffset(EKORDDataID::eTMovementType)));
            a_control->move_j_.setTrackingTypeValue(impl_->capture_frm_.move_J.getData<double>                  (impl_->capture_frm_.dfd_cmd_move_J.getOffset(EKORDDataID::eTMovementValue)));
            a_control->move_j_.setBlendType((TTP::BlendType)impl_->capture_frm_.move_J.getData<u_int8_t>        (impl_->capture_frm_.dfd_cmd_move_J.getOffset(EKORDDataID::eTBlendType)));
            a_control->move_j_.setBlendTypeValue(impl_->capture_frm_.move_J.getData<double>                     (impl_->capture_frm_.dfd_cmd_move_J.getOffset(EKORDDataID::eTBlendValue)));
            a_control->move_j_.setOverlayType((TTP::OverlayType)impl_->capture_frm_.move_J.getData<u_int8_t>    (impl_->capture_frm_.dfd_cmd_move_J.getOffset(EKORDDataID::eTOverlayType)));
            a_control->move_j_.setSyncTime(impl_->capture_frm_.move_J.getData<double>                           (impl_->capture_frm_.dfd_cmd_move_J.getOffset(EKORDDataID::eTSyncValue)));
            // place to calculate the statistics
            impl_->mux_stats_update.lock();
            stats_commands.calculateContinousMoveStats(impl_->capture_frm_.recent_captured_timestamp, impl_->capture_frm_.recent_captured_content_timestamp,
               a_control->move_j_);
            impl_->mux_stats_update.unlock();

            break;
        }

        case kr2::kord::protocol::EKORDItemID::eRequestStatusArm:{ 
            // here sequence number is different - always 0
            break;
        }

        case kr2::kord::protocol::EKORDItemID::eCommandMoveL:{
            if (!stats_commands.getAlarmState() && (controller.motion_flags_ & kr2::kord::protocol::EMotionFlags::MOTION_FLAG_SUSPENDED)){ // suspended 
                stats_commands.setAlarmState(true);
            }
            if (stats_commands.getAlarmState()){ // if alarm - ignore commands = clearing old capture frame
                impl_->capture_frm_.iid = EKORDItemID::eNone;
                break;
            }
            a_control->capture_ts_ = impl_->capture_frm_.rcv_cmd_linear_ts;
            a_control->sequence_num_ = impl_->capture_frm_.move_L.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_move_L.getOffset(EKORDDataID::eSequenceNumber));
            a_control->move_l_.setTargets(impl_->capture_frm_.move_L.getData<std::array<double, 6UL>>           (impl_->capture_frm_.dfd_cmd_move_L.getOffset(EKORDDataID::eFrmTCPPose)));
            a_control->move_l_.setTrackingType((TTP::TrackingType)impl_->capture_frm_.move_L.getData<u_int8_t>  (impl_->capture_frm_.dfd_cmd_move_L.getOffset(EKORDDataID::eTMovementType)));
            a_control->move_l_.setTrackingTypeValue(impl_->capture_frm_.move_L.getData<double>                  (impl_->capture_frm_.dfd_cmd_move_L.getOffset(EKORDDataID::eTMovementValue)));
            a_control->move_l_.setBlendType((TTP::BlendType)impl_->capture_frm_.move_L.getData<u_int8_t>        (impl_->capture_frm_.dfd_cmd_move_L.getOffset(EKORDDataID::eTBlendType)));
            a_control->move_l_.setBlendTypeValue(impl_->capture_frm_.move_L.getData<double>                     (impl_->capture_frm_.dfd_cmd_move_L.getOffset(EKORDDataID::eTBlendValue)));
            a_control->move_l_.setOverlayType((TTP::OverlayType)impl_->capture_frm_.move_L.getData<u_int8_t>    (impl_->capture_frm_.dfd_cmd_move_L.getOffset(EKORDDataID::eTOverlayType)));
            a_control->move_l_.setSyncTime(impl_->capture_frm_.move_L.getData<double>                           (impl_->capture_frm_.dfd_cmd_move_L.getOffset(EKORDDataID::eTSyncValue)));
            // place to calculate the statistics
            impl_->mux_stats_update.lock();
            stats_commands.calculateContinousMoveStats(impl_->capture_frm_.recent_captured_timestamp, impl_->capture_frm_.recent_captured_content_timestamp,
               a_control->move_l_);
            impl_->mux_stats_update.unlock();
            
            break;
        }

        case kr2::kord::protocol::EKORDItemID::eCommandJointFirmware:{
            
            
            unsigned int j = 0;
            int64_t frame_stamp_ = impl_->capture_frm_.cmd_jfw.getData<int64_t>(impl_->capture_frm_.dfd_cmd_jfw.getOffset(EKORDDataID::eTxStamp));
            
            // If the frame has been fetched before, do not fetch again.
            // It would overwrite the current fw commands and would not process all of them.
            if (a_control->jnt_fw_cmd_frame_stamp_ == frame_stamp_) break;

            a_control->sequence_num_ = impl_->capture_frm_.cmd_jfw.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_jfw.getOffset(EKORDDataID::eSequenceNumber));
            a_control->capture_ts_ = impl_->capture_frm_.rcv_cmd_jfw_ts;
            a_control->jnt_fw_cmd_frame_stamp_ = frame_stamp_;
            std::array<unsigned int, 7UL> cmds =impl_->capture_frm_.cmd_jfw.getData<std::array<unsigned int, 7UL>>(impl_->capture_frm_.dfd_cmd_jfw.getOffset(EKORDDataID::eJControlCMD));
            if (a_control->jnt_fw_commands_.size() != 0){
                CLOG_INFO("Got overlaping frame with id: " << impl_->capture_frm_.cmd_jfw.getData<int64_t>      (impl_->capture_frm_.dfd_cmd_jfw.getOffset(kr2::kord::protocol::EKORDDataID::eTxStamp)));
                // the last command wasn't processed fully, bad case, we need to add new command to not processed and ignore this one (only first one is processed)
                a_control->failed_by_not_completed = impl_->capture_frm_.cmd_jfw.getData<int64_t>      (impl_->capture_frm_.dfd_cmd_jfw.getOffset(kr2::kord::protocol::EKORDDataID::eTxStamp));
                break;
            }
            CLOG_INFO("Went to update array with token: " << impl_->capture_frm_.cmd_jfw.getData<int64_t>      (impl_->capture_frm_.dfd_cmd_jfw.getOffset(kr2::kord::protocol::EKORDDataID::eTxStamp)));
            for (auto item : cmds){
                EJointFirmwareCommand cmd = static_cast<EJointFirmwareCommand>(item);
                if (cmd != EJointFirmwareCommand::eNone){
                    a_control->jnt_fw_commands_.push_back(std::make_pair(j, cmd));
                }
                j++;
            }
            a_control->composed_dependent_ = a_control->jnt_fw_commands_.size();
            a_control->time_stamp_token_ = impl_->capture_frm_.cmd_jfw.getData<int64_t>      (impl_->capture_frm_.dfd_cmd_jfw.getOffset(kr2::kord::protocol::EKORDDataID::eTxStamp));
            a_control->crc_error_ = !impl_->capture_frm_.checkCRC(impl_->capture_frm_.cmd_jfw.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_jfw.getOffset(kr2::kord::protocol::EKORDDataID::eCRCValue)), impl_->capture_frm_.cmd_jfw);

            break;
        }

        case kr2::kord::protocol::EKORDItemID::eStatusEchoResponse:{
            // here sequence number is not present

            // If the frame was already fetched, do not fetch it second time.
            if (a_control->capture_ts_ == impl_->capture_frm_.rcv_resp_status_ts) { break; }
            // CLOG_INFO("Fetch Item Status Echo accepted");

            // New frame was captured, fetch new frame.
            a_control->capture_ts_    = impl_->capture_frm_.rcv_resp_status_ts;
            a_control->status_echo_tx = impl_->capture_frm_.resp_status.getData<int64_t>(impl_->capture_frm_.dfd_resp_status.getOffset(EKORDDataID::eTxStamp));
            a_control->status_echo_rx = impl_->capture_frm_.resp_status.getData<int64_t>(impl_->capture_frm_.dfd_resp_status.getOffset(EKORDDataID::eRxStamp));
            break;
        }

        case kr2::kord::protocol::EKORDItemID::eRequestSystem:{
            // here sequence number is different

            // If the frame was already fetched, do not fetch it second time.
            if (a_control->capture_ts_ == impl_->capture_frm_.rcv_req_sys_ts) { break; }
            CLOG_INFO("Fetch Item Request System accepted");

            // New frame was captured, fetch new frame.
            a_control->capture_ts_    = impl_->capture_frm_.rcv_req_sys_ts;
            a_control->req_sys_tx     = impl_->capture_frm_.req_system.getData<int64_t>(impl_->capture_frm_.dfd_req_system.getOffset(EKORDDataID::eTxStamp));
            a_control->req_system_id_ = static_cast<kr2::kord::protocol::EControlCommandItems>(impl_->capture_frm_.req_system.getData<uint16_t>(impl_->capture_frm_.dfd_req_system.getOffset(EKORDDataID::eCTRCommandItem)));
            break;
        }


        case kr2::kord::protocol::EKORDItemID::eRequestTransfer:{
            // here sequence number is different

            // If the frame was already fetched, do not fetch it second time.
            if (a_control->capture_ts_ == impl_->capture_frm_.rcv_req_transf_ts) { break; }
            CLOG_INFO("Fetch Item Request Transfer accepted");

            // New frame was captured, fetch new frame.
            a_control->capture_ts_    = impl_->capture_frm_.rcv_req_transf_ts;
            a_control->req_sys_tx     = impl_->capture_frm_.req_transfer.getData<int64_t>(impl_->capture_frm_.dfd_req_transfer.getOffset(EKORDDataID::eTxStamp));
            a_control->req_system_id_ = static_cast<kr2::kord::protocol::EControlCommandItems>(impl_->capture_frm_.req_transfer.getData<uint16_t>(impl_->capture_frm_.dfd_req_transfer.getOffset(EKORDDataID::eCTRCommandItem)));
            CLOG_INFO("req_system_id_: " <<  impl_->capture_frm_.req_transfer.getData<uint16_t>(impl_->capture_frm_.dfd_req_transfer.getOffset(EKORDDataID::eCTRCommandItem)));

            a_control->transfer_mask_ = impl_->capture_frm_.req_transfer.getData<uint32_t>(impl_->capture_frm_.dfd_req_transfer.getOffset(EKORDDataID::eTranferMask));
            break;
        }

        // eCommandSetIOAnalogOut
        // case kr2::kord::protocol::EKORDItemID::eCommandSetIOAnalogOut:
        case kr2::kord::protocol::EKORDItemID::eCommandSetIODigitalOut:{

            if (a_control->capture_ts_ == impl_->capture_frm_.rcv_req_io_ts) { break; }
            CLOG_INFO("Fetch Item Request IOSet.");

            // New frame was captured, fetch new frame.

            a_control->sequence_num_ = impl_->capture_frm_.req_io.getData<uint16_t>(impl_->capture_frm_.dfd_req_io.getOffset(EKORDDataID::eSequenceNumber));
            a_control->capture_ts_    = impl_->capture_frm_.rcv_req_io_ts;
            a_control->req_sys_tx     = impl_->capture_frm_.req_io.getData<int64_t>(impl_->capture_frm_.dfd_req_io.getOffset(EKORDDataID::eTxStamp));
            a_control->req_system_id_ = static_cast<kr2::kord::protocol::EControlCommandItems>(impl_->capture_frm_.req_io.getData<uint16_t>(impl_->capture_frm_.dfd_req_io.getOffset(EKORDDataID::eCTRCommandItem)));

            a_control->v_    = impl_->capture_frm_.req_io.getData<uint8_t>(impl_->capture_frm_.dfd_req_io.getOffset(EKORDDataID::eIODigitalValue));
            a_control->mask_ = impl_->capture_frm_.req_io.getData<int64_t>(impl_->capture_frm_.dfd_req_io.getOffset(EKORDDataID::eIODigitalOutput));
            break;
        }

        case kr2::kord::protocol::EKORDItemID::eCommandMoveDirect:{

            a_control->sequence_num_ = impl_->capture_frm_.move_D.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_move_D.getOffset(EKORDDataID::eSequenceNumber));
            a_control->capture_ts_ = impl_->capture_frm_.rcv_cmd_direct_ts;
            a_control->move_j_.setTargets(impl_->capture_frm_.move_D.getData<std::array<double, 7UL>>      (impl_->capture_frm_.dfd_cmd_move_D.getOffset(kr2::kord::protocol::EKORDDataID::eJConfigurationArm)));
            a_control->move_j_.setTargetsD(impl_->capture_frm_.move_D.getData<std::array<double, 7UL>>     (impl_->capture_frm_.dfd_cmd_move_D.getOffset(kr2::kord::protocol::EKORDDataID::eJSpeedArm)));
            a_control->move_j_.setTargetsDD(impl_->capture_frm_.move_D.getData<std::array<double, 7UL>>    (impl_->capture_frm_.dfd_cmd_move_D.getOffset(kr2::kord::protocol::EKORDDataID::eJAccelerationArm)));
            a_control->move_j_.setTargetsTorque(impl_->capture_frm_.move_D.getData<std::array<double, 7UL>>(impl_->capture_frm_.dfd_cmd_move_D.getOffset(kr2::kord::protocol::EKORDDataID::eJTorqueArm)));
            break;
        }

        case kr2::kord::protocol::EKORDItemID::eCommandMoveVelocity:{

            a_control->sequence_num_ = impl_->capture_frm_.move_V.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_move_V.getOffset(EKORDDataID::eSequenceNumber));
            a_control->capture_ts_ = impl_->capture_frm_.rcv_cmd_velocity_ts;
            a_control->move_l_.setTargets(impl_->capture_frm_.move_V.getData<std::array<double, 6UL>>      (impl_->capture_frm_.dfd_cmd_move_V.getOffset(kr2::kord::protocol::EKORDDataID::eFrmTCPVelocity)));
            a_control->move_l_.setSyncTime(impl_->capture_frm_.move_V.getData<long>                  (impl_->capture_frm_.dfd_cmd_move_V.getOffset(EKORDDataID::eVelMoveSync)));
            a_control->move_l_.setPeriod(impl_->capture_frm_.move_V.getData<double>                  (impl_->capture_frm_.dfd_cmd_move_V.getOffset(EKORDDataID::eVelMovePeriod)));
            a_control->move_l_.setTimeout(impl_->capture_frm_.move_V.getData<double>                  (impl_->capture_frm_.dfd_cmd_move_V.getOffset(EKORDDataID::eVelMoveTimeout)));
            break;
        }

        case kr2::kord::protocol::EKORDItemID::eCommandSetFrame:{

            a_control->sequence_num_ = impl_->capture_frm_.set_Frame.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_set_Frame.getOffset(EKORDDataID::eSequenceNumber));
            a_control->capture_ts_ = impl_->capture_frm_.rcv_cmd_setFrame_ts;
            a_control->time_stamp_token_ = impl_->capture_frm_.set_Frame.getData<int64_t>      (impl_->capture_frm_.dfd_cmd_set_Frame.getOffset(kr2::kord::protocol::EKORDDataID::eTxStamp));
            a_control->pose_ = impl_->capture_frm_.set_Frame.getData<std::array<double, 6UL>>      (impl_->capture_frm_.dfd_cmd_set_Frame.getOffset(kr2::kord::protocol::EKORDDataID::eCTRSetFramePose));
            a_control->frame_id_ = impl_->capture_frm_.set_Frame.getData<uint8_t>      (impl_->capture_frm_.dfd_cmd_set_Frame.getOffset(kr2::kord::protocol::EKORDDataID::eCTRSetFrameId));
            a_control->ref_frame_ = impl_->capture_frm_.set_Frame.getData<uint8_t>      (impl_->capture_frm_.dfd_cmd_set_Frame.getOffset(kr2::kord::protocol::EKORDDataID::eCTRSetFrameRef));
            a_control->crc_error_ = !impl_->capture_frm_.checkCRC(impl_->capture_frm_.set_Frame.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_set_Frame.getOffset(kr2::kord::protocol::EKORDDataID::eCRCValue)), impl_->capture_frm_.set_Frame);

            break;
        }


        case kr2::kord::protocol::EKORDItemID::eCommandSetLoad:{
            a_control->sequence_num_ = impl_->capture_frm_.set_Load.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_set_Load.getOffset(EKORDDataID::eSequenceNumber));
            
            a_control->capture_ts_ = impl_->capture_frm_.rcv_cmd_setLoad_ts;

            a_control->time_stamp_token_ = impl_->capture_frm_.set_Load.getData<int64_t>      (impl_->capture_frm_.dfd_cmd_set_Load.getOffset(kr2::kord::protocol::EKORDDataID::eTxStamp));
            a_control->load_cog_ = impl_->capture_frm_.set_Load.getData<std::array<double, 3UL>>      (impl_->capture_frm_.dfd_cmd_set_Load.getOffset(kr2::kord::protocol::EKORDDataID::eCTRSetLoadCoG));
            a_control->load_inertia_ = impl_->capture_frm_.set_Load.getData<std::array<double, 6UL>>      (impl_->capture_frm_.dfd_cmd_set_Load.getOffset(kr2::kord::protocol::EKORDDataID::eCTRSetLoadInertia));
            a_control->load_mass_ = impl_->capture_frm_.set_Load.getData<double>      (impl_->capture_frm_.dfd_cmd_set_Load.getOffset(kr2::kord::protocol::EKORDDataID::eCTRSetLoadMass));
            a_control->load_id_ = impl_->capture_frm_.set_Load.getData<uint8_t>      (impl_->capture_frm_.dfd_cmd_set_Load.getOffset(kr2::kord::protocol::EKORDDataID::eCTRSetLoadId));

            a_control->crc_error_ = !impl_->capture_frm_.checkCRC(impl_->capture_frm_.set_Load.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_set_Load.getOffset(kr2::kord::protocol::EKORDDataID::eCRCValue)), impl_->capture_frm_.set_Load);
            break;
        }

        case kr2::kord::protocol::EKORDItemID::eCommandCleanAlarm:{
            a_control->sequence_num_ = impl_->capture_frm_.clean_Alarm.getData<uint16_t>(impl_->capture_frm_.dfd_cmd_clean_Alarm.getOffset(EKORDDataID::eSequenceNumber));
            
            a_control->capture_ts_ = impl_->capture_frm_.rcv_cmd_cleanAlarm_ts;

            a_control->time_stamp_token_ = impl_->capture_frm_.clean_Alarm.getData<int64_t>      (impl_->capture_frm_.dfd_cmd_clean_Alarm.getOffset(kr2::kord::protocol::EKORDDataID::eTxStamp));
            a_control->alarm_id_ = impl_->capture_frm_.clean_Alarm.getData<uint8_t>      (impl_->capture_frm_.dfd_cmd_clean_Alarm.getOffset(kr2::kord::protocol::EKORDDataID::eCTRCleanAlarmID));
            break;
        }



        case kr2::kord::protocol::EKORDItemID::eNone:
        default:
        {
            // nothing to do
            impl_->mux_capture_.unlock();
            return false;
        }
    }
    
    impl_->mux_capture_.unlock();

    impl_->mux_stats_update.lock();
    stats_commands.processNewSequenceNumber(a_control->sequence_num_);
    impl_->mux_stats_update.unlock();

    // if the frame was captured after a_recent tick, then it is considered recent enough
    auto elapsed_from_tick = std::chrono::duration_cast<std::chrono::microseconds>(a_control->capture_ts_ - latest_ok);
    //CLOG_INFO("elapsed_from_tick: " << elapsed_from_tick.count() << '\n');

    return elapsed_from_tick.count() > 0;
}

void kr2::kord::KORDServer::updateStatus(const kr2::kord::RobotModel &a_model, const RobotController &a_controller, StatusBuffer<CommandStatus> &a_command_statuses)
{
    impl_->mux_transmit_.lock();
    // std::array<double, 7UL> ar;
    // std::array<double, 6UL> ar6;

    // Clean the buffer before the use
    impl_->transmit_frm_.buffer.clear();
    
    impl_->updateVectorData(EKORDDataID::eJConfigurationArm,    a_model.joint_position_);
    impl_->updateVectorData(EKORDDataID::eJSpeedArm,            a_model.joint_speed_);
    impl_->updateVectorData(EKORDDataID::eJAccelerationArm,     a_model.joint_acceleration_);
    impl_->updateVectorData(EKORDDataID::eJTorqueArm,           a_model.joint_torque_);

    if (a_model.sensors_.size() == 7){
        impl_->updateArrayData(EKORDDataID::eJTemperatureBoard, a_model.getJointsTemperature(0));
        impl_->updateArrayData(EKORDDataID::eJTemperatureJointEncoder, a_model.getJointsTemperature(1));
        impl_->updateArrayData(EKORDDataID::eJTemperatureRotorEncoder, a_model.getJointsTemperature(2));
        impl_->updateArrayData(EKORDDataID::eJErrorBits,        a_model.getJointsErrorBits());
        impl_->updateArrayData(EKORDDataID::eJStatusBits,       a_model.getJointsStatusBits());
        impl_->updateArrayData(EKORDDataID::eJStatorVoltage,    a_model.getJointsStatorVoltage());
        impl_->updateArrayData(EKORDDataID::eJStatorCurrent,    a_model.getJointsStatorCurrent());
        impl_->updateArrayData(EKORDDataID::eJTorqueModel,    a_model.getJointsTorques());
    }
    // CPU state
    impl_->transmit_frm_.status_builder.addData(a_controller.cpu_state_[impl_->cpu_state_index].id_,    impl_->status_dfd.getOffset(EKORDDataID::eCPUStateId));
    impl_->transmit_frm_.status_builder.addData(temperatureEncoder(a_controller.cpu_state_[impl_->cpu_state_index].value_),  impl_->status_dfd.getOffset(EKORDDataID::eCPUStateVal));
    impl_->cpu_state_index = (impl_->cpu_state_index + 1) % num_cpu_states;

    //IO Board temperature
    impl_->transmit_frm_.status_builder.addData(a_model.getIOBoardTemperature(),    impl_->status_dfd.getOffset(EKORDDataID::eIOBCabinetTemperature));
    //CLOG_INFO( "Temperature = " << a_model.getIOBoardTemperature());

    // Frames
    impl_->transmit_frm_.status_builder.addData(a_model.frames_[impl_->frame_index_].id_,    impl_->status_dfd.getOffset(EKORDDataID::eFrameId));
    impl_->transmit_frm_.status_builder.addData(a_model.frames_[impl_->frame_index_].ref_value_id_,  impl_->status_dfd.getOffset(EKORDDataID::eFramePoseRef));
    impl_->updateVectorData(EKORDDataID::eFramePose,  a_model.frames_[impl_->frame_index_].pose_);
    impl_->frame_index_ = (impl_->frame_index_ + 1) % num_frames_values;

    // Loads
    impl_->transmit_frm_.status_builder.addData(a_model.loads_[impl_->load_index_].id_,    impl_->status_dfd.getOffset(EKORDDataID::eLoadId));
    impl_->updateVectorData(EKORDDataID::eLoadCoG,  a_model.loads_[impl_->load_index_].cog_);
    impl_->updateVectorData(EKORDDataID::eLoadPose,  a_model.loads_[impl_->load_index_].pose_);
    impl_->updateVectorData(EKORDDataID::eLoadInertia,  a_model.loads_[impl_->load_index_].inertia_);
    impl_->transmit_frm_.status_builder.addData(a_model.loads_[impl_->load_index_].mass_,    impl_->status_dfd.getOffset(EKORDDataID::eLoadMass));
    impl_->load_index_ = (impl_->load_index_ + 1) % num_loads;

    // TCP model
    impl_->updateVectorData(EKORDDataID::eFrmTCPPose,  a_model.tcp_model_);
    // TFC model
    impl_->updateVectorData(EKORDDataID::eFrmTFCModel, a_model.tcp_model_fwk_);

    impl_->transmit_frm_.status_builder.addData(a_controller.rc_safety_flags_, impl_->status_dfd.getOffset(EKORDDataID::eRCSafetyFlag));
    impl_->transmit_frm_.status_builder.addData(a_controller.safety_mode_,     impl_->status_dfd.getOffset(EKORDDataID::eRCSafetyMode));
    impl_->transmit_frm_.status_builder.addData(a_controller.master_speed_,    impl_->status_dfd.getOffset(EKORDDataID::eRCMasterSpeed));
    impl_->transmit_frm_.status_builder.addData(a_controller.motion_flags_,    impl_->status_dfd.getOffset(EKORDDataID::eRCMotionFlags));
    impl_->transmit_frm_.status_builder.addData(a_controller.rc_hw_flags_,     impl_->status_dfd.getOffset(EKORDDataID::eRCHWFlags));
    impl_->transmit_frm_.status_builder.addData(a_controller.rc_button_flags_, impl_->status_dfd.getOffset(EKORDDataID::eRCButtonFlags));
    impl_->transmit_frm_.status_builder.addData(a_controller.rc_system_alarm_state_, impl_->status_dfd.getOffset(EKORDDataID::eRCSystemAlarmState));

    // Digital Register
    impl_->transmit_frm_.status_builder.addData<RobotIO::IORegister>(a_model.io_model_.getInputRegister(), impl_->status_dfd.getOffset(EKORDDataID::eIODigitalInput));
    // impl_->transmit_frm_.status_builder.addData<RobotIO::Analog>(a_model.io_model_.getInputAnalog(),    impl_->status_dfd.getOffset(EKORDDataID::eIOAnalogInput));
    // Analog
    impl_->transmit_frm_.status_builder.addData<RobotIO::IORegister>(a_model.io_model_.getOutputRegister(), impl_->status_dfd.getOffset(EKORDDataID::eIODigitalOutput));
    // impl_->transmit_frm_.status_builder.addData<RobotIO::Analog>(a_model.io_model_.getOutputAnalog(),    impl_->status_dfd.getOffset(EKORDDataID::eIOAnalogOutput));

    impl_->transmit_frm_.status_builder.addData<uint16_t>(a_controller.req_system_id_,         impl_->status_dfd.getOffset(EKORDDataID::eCTRCommandItem));
    impl_->transmit_frm_.status_builder.addData<int64_t> (a_controller.req_sys_tx,             impl_->status_dfd.getOffset(EKORDDataID::eCTRCommandTS));
    impl_->transmit_frm_.status_builder.addData<uint16_t>(a_controller.request_system_status_, impl_->status_dfd.getOffset(EKORDDataID::eCTRCommandStatus));
    impl_->transmit_frm_.status_builder.addData<uint32_t>(a_controller.request_error_,         impl_->status_dfd.getOffset(EKORDDataID::eCTRCommandErrorCode));

    // Convert the system events to the byte array and pass it to the builder.
    for (auto event :  a_controller.system_events_.events_){
        event.toByteArray(impl_->transmit_frm_.buffer);
    }
    impl_->transmit_frm_.status_builder.addData(impl_->transmit_frm_.buffer, impl_->status_dfd.getOffset(EKORDDataID::eEventsArray));

    // Last command status
    CommandStatus last_stat;
    if (a_command_statuses.getRecentStatusAndDelete(last_stat)){
        CLOG_INFO("Last token "  << last_stat.token);

        impl_->transmit_frm_.status_builder.addData<uint64_t>(last_stat.token,         impl_->status_dfd.getOffset(EKORDDataID::eLastCommandToken));
        impl_->transmit_frm_.status_builder.addData<int8_t>(last_stat.error_code,      impl_->status_dfd.getOffset(EKORDDataID::eLastCommandErrorCode));
    }

    impl_->mux_transmit_.unlock();
}

bool kr2::kord::KORDServer::addResponse(kr2::kord::RobotResponseItem *)
{
    // TODO: add a arbitrary frame buffering
    return true;
}

bool kr2::kord::KORDServer::dispatch(int64_t& sent_status_tx_stamp, CommandsMonitor& stats_commands)
{
    impl_->updateArmStatusStamps(sent_status_tx_stamp, stats_commands);
    impl_->updateKordFrame();
    int n = impl_->transmitFrame();

    // TODO: Transmit content responses if present
    impl_->mux_transmit_.lock();
    impl_->transmit_frm_.status_builder.clear();
    // transmit_frm_.content_builder.clear();
    impl_->mux_transmit_.unlock();


    return (n > 0);
}
